package se.umu.jobsubmission.impl.jobsubmissionservice; 

import java.io.StringWriter;
import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;
import javax.xml.rpc.server.ServiceLifecycle;

import org.apache.axis.MessageContext;
import org.apache.log4j.Logger;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.ContextUtils;

import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.RequestJob;
import se.umu.jobsubmission.RequestJobResponse;
import se.umu.jobsubmission.SubmittedJob;
import se.umu.jobsubmission.impl.submitter.Submitter;

import org.globus.wsrf.utils.FaultHelper; 
import se.umu.jobsubmission.JobRequestFaultType; 
import se.umu.jobsubmission.NoResourcesFoundFaultType; 

import	org.oasis.wsrf.faults.BaseFaultType;
import	se.umu.jobsubmission.NoResourcesFoundFaultType; 
import 	se.umu.jobsubmission.InvalidJobDescriptionFaultType; 
import	se.umu.jobsubmission.JobRequestFaultType;
import 	se.umu.jobsubmission.NoAptResourcesFoundFaultType;


import org.oasis.wsrf.faults.BaseFaultType; 

import org.globus.wsrf.encoding.ObjectSerializer;

import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;

/**
 * Service for brokering (scheduling) of Grid jobs. 
 * Service recieves job requests containing JSDL job description, 
 * list of custom index servers to query for resources, 
 * and optionally, a document describing job preferences. 
 */

public class JobSubmissionService implements ServiceLifecycle {

    private static Logger logger = Logger.getLogger(JobSubmissionService.class); 
    private Submitter submitter; 
    

    public JobSubmissionService() {
    }

    // Specified by ServiceLifecycle
    public void destroy() {
        logger.debug("Destroy invoked"); 
        submitter.shutdown(); 
    }
    
    // Invoked upon creation of service, initiates 
    // xml driver and configures submitter. 
    public void init(java.lang.Object context) throws ServiceException {

        try {

            MessageContext ctx = MessageContext.getCurrentContext(); 

            // Configure submitter
            String configFile = (String) ContextUtils.getServiceProperty(ctx, 
                    JobSubmissionConstants.SUBMITTER_CONFIG_FILE); 
            String globusLocation = ContainerConfig.getGlobusLocation(); 
            String filePath = globusLocation + configFile; 

            logger.debug("JSS config file: " + filePath); 
            submitter = new Submitter(filePath); 
            
        } catch (Exception e) {
            logger.error(e.getMessage()); 
            throw new ServiceException("Error configuring service", e); 
        }
        
    }
    
  
    /**
     * Method specified by wsdl porttype
     * @param param Contains the job description and optionally, 
     * job preferences and index servers to query for resources (instead
     * of using the default ones). 
     * @return Job ID as a string (can be EPR or other kind of URI)
     * @throws RemoteException
     * @throws JobRequestFaultType
     */
    public RequestJobResponse requestJob(RequestJob param) 
    	throws RemoteException, 
		BaseFaultType, 
		NoResourcesFoundFaultType, 
		InvalidJobDescriptionFaultType, 
		JobRequestFaultType, 
		NoAptResourcesFoundFaultType { 

        logger.info("New job requested"); 
    	
        SubmittedJob jobInfo = null; 
        
        JobDefinition_Type jobDesc = null; 
        
        // Submit job. 
        try {
        	
            JobRequest request = param.getRequest(); 
            if(request == null) 
                throw new RemoteException("No job request specified!"); 
            
            String host = request.getClientURL(); 
            logger.info("Job submitted from host: " + host); 

            jobDesc = request.getJobDescription(); 
            
            Boolean clientStages = request.getClientStagesFiles(); 
            boolean clientStagesInput = true; 
            if(clientStages != null && clientStages.booleanValue() == false) 
            	clientStagesInput = false; 
            
            String IndexServerURIs[] = request.getIndexServerURIs(); 
            JobPreferencesType jobPref = request.getJobPreferences(); 
            String clientURL = request.getClientURL(); 
            
            jobInfo = submitter.submitJob(jobDesc, 
            		IndexServerURIs, 
					jobPref, 
					clientURL, 
					clientStagesInput); 
          
        } 
        catch (RemoteException re) {
        	logger.error("Job request failed (RemoteException): " 
        			+ re.getMessage());         	
        	throw re; 
        }
        catch (Exception e) {
        	logger.error("Job request failed: " + e.getMessage()); 
        	throw new RemoteException(e.getMessage()); 
            
        }
     
        // Get resource home and create new resource. 
    	ResourceContext ctx = null;
		JobSubmissionHome home = null;
		
		// Add new resource property, synchronize this as 
		// more than one job may be submitted simultaneously
		// FIXME: Enought to synchronize the home.create(jobInfo) call? 
		// Maybe, but this is not a bottleneck as far as I understand...
		synchronized (this) {
			try {			
				ctx = ResourceContext.getResourceContext();
				home = (JobSubmissionHome) ctx.getResourceHome();			
				home.create(jobInfo); 						
			} catch (Exception e) {
				logger.error("Could not create resource"); 
				throw new RemoteException("Could not create resource", e); 
			}
		}

		logger.info("Job submission done, returning job info"); 
        return new RequestJobResponse(jobInfo.getLocalID()); 
    }
}

